@@ -31,6 +31,7 @@ _SECTION_HEADER_FORMAT = (
     'I'   # Characteristics
 )
 _SECTION_HEADER_SIZE = struct.calcsize(_SECTION_HEADER_FORMAT)
+_NEW_SECTION_CHARACTERISTICS = pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_READ'] | pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_CNT_INITIALIZED_DATA']
 
 
 class PEContext:
@@ -270,10 +271,10 @@ def get_direct_needed(lib_path: str, include_delay_imports: bool, lower: bool, v
     return needed
 
 
-def get_direct_mangleable_needed(lib_path: str, no_dlls: set, no_mangles: set, verbose: int) -> set:
-    """Given the path to a shared library, return a set containing the
-    lowercase DLL names of all direct dependencies that belong in the wheel and
-    should be name-mangled.
+def get_direct_mangleable_needed(lib_path: str, no_dlls: set, no_mangles: set, verbose: int) -> list:
+    """Given the path to a shared library, return a deterministically-ordered
+    list containing the lowercase DLL names of all direct dependencies that
+    belong in the wheel and should be name-mangled.
 
     no_dlls is a set of lowercase additional DLL names that do not belong in
     the wheel.
@@ -284,7 +285,7 @@ def get_direct_mangleable_needed(lib_path: str, no_dlls: set, no_mangles: set, v
         for attr in ('DIRECTORY_ENTRY_IMPORT', 'DIRECTORY_ENTRY_DELAY_IMPORT'):
             if hasattr(pe, attr):
                 imports = itertools.chain(imports, getattr(pe, attr))
-        needed = set()
+        needed = []
         lib_arch = MachineType.machine_field_to_type(pe.FILE_HEADER.Machine)
         if not lib_arch:
             raise ValueError(f'{lib_path} has an unsupported CPU architecture')
@@ -296,7 +297,7 @@ def get_direct_mangleable_needed(lib_path: str, no_dlls: set, no_mangles: set, v
                     not any(r.fullmatch(dll_name) for r in _dll_list.ignore_regexes) and \
                     dll_name not in no_mangles and \
                     not any(r.fullmatch(dll_name) for r in _dll_list.no_mangle_regexes):
-                needed.add(dll_name)
+                needed.append(dll_name)
     return needed
 
 
@@ -367,53 +368,111 @@ def _round_to_next(size: int, alignment: int) -> int:
         return alignment * (size // alignment + 1)
 
 
-def replace_needed(lib_path: str, old_deps: typing.Iterable[str], name_map: typing.Dict[str, str], strip: bool, verbose: int) -> None:
+def _are_characteristics_suitable(section: pefile.SectionStructure) -> bool:
+    """Determine whether a PE section is suitable for writing new DLL names."""
+    return (not section.IMAGE_SCN_CNT_CODE and
+            section.IMAGE_SCN_CNT_INITIALIZED_DATA and
+            not section.IMAGE_SCN_CNT_UNINITIALIZED_DATA and
+            not section.IMAGE_SCN_GPREL and
+            not section.IMAGE_SCN_LNK_NRELOC_OVFL and
+            not section.IMAGE_SCN_MEM_DISCARDABLE and
+            not section.IMAGE_SCN_MEM_EXECUTE and
+            section.IMAGE_SCN_MEM_READ)
+
+
+def _get_pe_size_and_enough_padding(pe: pefile.PE, new_dlls: typing.Iterable[bytes]) -> typing.Tuple[int, bool]:
+    """Determine the size of a PE file (excluding any trailing data) and
+    whether the file has enough padding for writing the elements of new_dlls.
+
+    Determining whether the file has enough padding is an instance of the NP-
+    complete bin packing problem, so we use the Next Fit approximation
+    algorithm. new_dlls must be in a deterministic order in order for the Next
+    Fit algorithm to return a deterministic result.
+
+    Precondition: pe must have at least 1 DLL dependency"""
+    pe_size = 0
+    new_dlls = list(new_dlls)
+    dlls_i = 0
+    pe.sections.sort(key=lambda section: section.VirtualAddress)
+    for section_i, section in enumerate(pe.sections):
+        pe_size = max(pe_size, section.PointerToRawData + section.SizeOfRawData)
+        if dlls_i < len(new_dlls) and section.Misc_VirtualSize < section.SizeOfRawData and _are_characteristics_suitable(section):
+            padding_size = section.SizeOfRawData - section.Misc_VirtualSize
+            if section_i + 1 < len(pe.sections):
+                padding_size = min(padding_size, pe.sections[section_i + 1].VirtualAddress - section.VirtualAddress - section.Misc_VirtualSize)
+            while dlls_i < len(new_dlls) and len(new_dlls[dlls_i]) < padding_size:
+                padding_size -= len(new_dlls[dlls_i]) + 1
+                dlls_i += 1
+    return pe_size, dlls_i == len(new_dlls)
+
+
+def replace_needed(lib_path: str, old_deps: typing.List[str], name_map: typing.Dict[str, str], strip: bool, verbose: int, test: typing.List[str]) -> None:
     """For the DLL at lib_path, replace its declared dependencies on old_deps
     with those in name_map.
-    old_deps: a subset of the dependencies that lib_path has
+    old_deps: a subset of the dependencies that lib_path has, in list form
     name_map: a dict that maps an old dependency name to a new name, must
         contain at least all the keys in old_deps
-    strip: whether to try to strip DLLs that contain trailing data"""
-    used_name_map = {dep.lower().encode('utf-8'): name_map[dep].encode('utf-8') for dep in old_deps}
-        # map from lowercase old DLL name in bytes to new DLL name in bytes, includes only those DLLs that will be mangled
-    if not used_name_map:
+    strip: whether to try to strip DLLs that contain trailing data if not
+        enough internal padding exists"""
+    if not old_deps:
         # no dependency names to change
         return
+    name_map = {dep.lower().encode('utf-8'): name_map[dep].encode('utf-8') for dep in old_deps}
+        # keep only the DLLs that will be mangled
+
+    # New dependency names are longer than the old ones, so we cannot simply
+    # overwrite the bytes of the old dependency names. Determine whether the PE
+    # file has enough usable internal padding to use to write the new
+    # dependency names. If so, overwrite the padding. Otherwise, as long as the
+    # PE file contains no trailing data or the trailing data can be stripped,
+    # append a new PE section to store the new names. Determining whether
+    # enough internal padding exists is an instance of the bin packing problem
+    # in which the new dependency names are items and the contiguous padding
+    # runs are bins. The bin packing problem is NP-hard, so for simplicity, we
+    # use the Next Fit algorithm.
     with PEContext(lib_path, None, False, verbose) as pe:
-        pe_size = max(section.PointerToRawData + section.SizeOfRawData for section in pe.sections)
-    if pe_size < os.path.getsize(lib_path) and strip:
+        pe_size, enough_padding = _get_pe_size_and_enough_padding(pe, name_map.values())
+    if 'not_enough_padding' in test:
+        enough_padding = False
+    if not enough_padding and pe_size < os.path.getsize(lib_path) and strip:
         # try to strip the trailing data
         try:
             subprocess.check_call(['strip', '-s', lib_path])
         except FileNotFoundError:
             raise FileNotFoundError('GNU strip not found in PATH') from None
         with PEContext(lib_path, None, False, verbose) as pe:
-            pe_size = max(section.PointerToRawData + section.SizeOfRawData for section in pe.sections)
+            pe_size, enough_padding = _get_pe_size_and_enough_padding(pe, name_map.values())
+
     lib_name = os.path.basename(lib_path)
-    if pe_size < os.path.getsize(lib_path):
+    with open(lib_path, 'rb') as f:
+        # workaround for https://github.com/erocarrera/pefile/issues/356
+        lib_data = f.read()
+    if not enough_padding and pe_size < len(lib_data):
         # cannot rename dependencies due to trailing data
         if strip:
             raise RuntimeError(textwrap.fill(
                 f'Unable to rename the dependencies of {lib_name} because '
-                'this DLL contains trailing data after the point where the '
-                'DLL file specification ends. The GNU strip utility was run '
-                'automatically in attempt to remove the trailing data but '
-                'failed to remove all of it. Unless you have knowledge to the '
-                'contrary, you should assume that the trailing data exist for '
-                'an important reason and are not safe to remove. Include '
-                f'{os.pathsep.join(old_deps)} in the --no-mangle flag to fix '
-                'this error.',
+                'this DLL does not contain enough internal padding to fit the '
+                'new dependency names, and it contains trailing data after '
+                'the point where the DLL file specification ends. The GNU '
+                'strip utility was run automatically in attempt to remove the '
+                'trailing data but failed to remove all of it. Unless you '
+                'have knowledge to the contrary, you should assume that the '
+                'trailing data exist for an important reason and are not safe '
+                f'to remove. Include {os.pathsep.join(old_deps)} in the '
+                '--no-mangle flag to fix this error.',
                 initial_indent=' ' * len('RuntimeError: ')).lstrip())
         else:
             error_text = [
                 textwrap.fill(
                     f'Unable to rename the dependencies of {lib_name} because '
-                    'this DLL contains trailing data after the point where '
-                    'the DLL file specification ends. Commonly, the trailing '
-                    'data consist of symbols that can be safely removed, '
-                    'although there exist situations where the data must be '
-                    'present for the DLL to function properly. Here are your '
-                    'options.',
+                    'this DLL does not contain enough internal padding to fit '
+                    'the new dependency names, and it contains trailing data '
+                    'after the point where the DLL file specification ends. '
+                    'Commonly, the trailing data consist of symbols that can '
+                    'be safely removed, although there exist situations where '
+                    'the data must be present for the DLL to function '
+                    'properly. Here are your options.',
                     initial_indent=' ' * len('RuntimeError: ')).lstrip(),
                 '\n',
                 textwrap.fill(
@@ -436,82 +495,131 @@ def replace_needed(lib_path: str, old_deps: typing.Iterable[str], name_map: typi
             ]
             raise RuntimeError(''.join(error_text))
 
-    # generate section data containing strings for new DLL names
-    section_offset_mapping = {}  # map lowercase old DLL name to offset of new DLL name within new section
-    with io.BytesIO() as new_section_data:
-        for old_dll, new_dll in used_name_map.items():
-            section_offset_mapping[old_dll] = new_section_data.tell()
-            new_section_data.write(new_dll)
-            new_section_data.write(b'\x00')
-        new_section_data = new_section_data.getvalue()
+    with PEContext(None, lib_data, True, verbose) as pe:
+        if enough_padding:
+            # overwrite padding with new DLL names
+            pe.sections.sort(key=lambda section: section.VirtualAddress)
+            new_dlls_rva_mapping = {}  # map lowercase old DLL name to rva of new DLL name
+            old_new_dlls = list(name_map.items())
+            dlls_i = 0
+            for section_i, section in enumerate(pe.sections):
+                rva = section.VirtualAddress + section.Misc_VirtualSize
+                if (section.Misc_VirtualSize < section.SizeOfRawData and
+                        (section_i + 1 == len(pe.sections) or rva < pe.sections[section_i + 1].VirtualAddress)
+                        and _are_characteristics_suitable(section)):
+                    file_offset = section.PointerToRawData + section.Misc_VirtualSize
+                    while (dlls_i < len(old_new_dlls) and
+                           file_offset + len(old_new_dlls[dlls_i][1]) < section.PointerToRawData + section.SizeOfRawData and
+                           (section_i + 1 == len(pe.sections) or rva + len(old_new_dlls[dlls_i][1]) < pe.sections[section_i + 1].VirtualAddress)):
+                        old_dll, new_dll = old_new_dlls[dlls_i]
+                        pe.set_bytes_at_offset(file_offset, new_dll + b'\x00')
+                        new_dlls_rva_mapping[old_dll] = rva
+                        new_dll_size = len(new_dll) + 1
+                        section.Misc_VirtualSize += new_dll_size
+                        rva += new_dll_size
+                        file_offset += new_dll_size
+                        dlls_i += 1
+                    if dlls_i == len(old_new_dlls):
+                        break
+            else:
+                raise RuntimeError('Expected enough internal padding to write new DLL names but ran out of space when writing them')
 
-    # update PE headers and import tables to what they will need to be once the
-    # new section header and new section are added
-    with PEContext(lib_path, None, True, verbose) as pe:
-        section_table_end = pe.DOS_HEADER.e_lfanew + 4 + \
-                            pe.FILE_HEADER.sizeof() + \
-                            pe.FILE_HEADER.SizeOfOptionalHeader + \
-                            pe.FILE_HEADER.NumberOfSections * _SECTION_HEADER_SIZE
-        if pe.OPTIONAL_HEADER.SizeOfHeaders - section_table_end >= _SECTION_HEADER_SIZE:
-            # there's enough unused space to add new section header
-            new_section_header_space_needed = 0
-        else:
-            # there's not enough unused space to add new section header;
-            # calculate how much extra space to add
-            new_section_header_space_needed = _round_to_next(_SECTION_HEADER_SIZE, pe.OPTIONAL_HEADER.FileAlignment)
-        new_section_data_size = _round_to_next(len(new_section_data), pe.OPTIONAL_HEADER.FileAlignment)
-        new_section_rva = _round_to_next(max(section.VirtualAddress + section.Misc_VirtualSize for section in pe.sections), pe.OPTIONAL_HEADER.SectionAlignment)
+            # update headers and import tables
+            pe.OPTIONAL_HEADER.SizeOfInitializedData = sum(max(section.SizeOfRawData, _round_to_next(section.Misc_VirtualSize, pe.OPTIONAL_HEADER.FileAlignment)) for section in pe.sections if section.IMAGE_SCN_CNT_INITIALIZED_DATA)
+            pe.OPTIONAL_HEADER.SizeOfImage = _round_to_next(pe.sections[-1].VirtualAddress + pe.sections[-1].Misc_VirtualSize, pe.OPTIONAL_HEADER.SectionAlignment)
+            if hasattr(pe, 'DIRECTORY_ENTRY_IMPORT'):
+                for entry in pe.DIRECTORY_ENTRY_IMPORT:
+                    old_dll = entry.dll.lower()
+                    if old_dll in new_dlls_rva_mapping:
+                        entry.struct.Name = new_dlls_rva_mapping[old_dll]
+            if hasattr(pe, 'DIRECTORY_ENTRY_DELAY_IMPORT'):
+                for entry in pe.DIRECTORY_ENTRY_DELAY_IMPORT:
+                    old_dll = entry.dll.lower()
+                    if old_dll in new_dlls_rva_mapping:
+                        entry.struct.szName = new_dlls_rva_mapping[old_dll]
+        else:  # not enough_padding
+            # generate data containing strings for new DLL names
+            new_section_offset_mapping = {}  # map lowercase old DLL name to offset of new DLL name within new section
+            with io.BytesIO() as new_section_data:
+                for old_dll, new_dll in name_map.items():
+                    new_section_offset_mapping[old_dll] = new_section_data.tell()
+                    new_section_data.write(new_dll)
+                    new_section_data.write(b'\x00')
+                new_section_data = new_section_data.getvalue()
 
-        pe.FILE_HEADER.NumberOfSections += 1
-        pe.OPTIONAL_HEADER.SizeOfInitializedData += len(new_section_data)
-        pe.OPTIONAL_HEADER.SizeOfImage += new_section_header_space_needed + new_section_data_size
-        pe.OPTIONAL_HEADER.SizeOfHeaders += new_section_header_space_needed
-        for section in pe.sections:
-            section.PointerToRawData += new_section_header_space_needed
+            # update PE headers to what they will need to be once the
+            # new section header and new section are added
+            section_table_end = (pe.DOS_HEADER.e_lfanew + 4 +
+                                 pe.FILE_HEADER.sizeof() +
+                                 pe.FILE_HEADER.SizeOfOptionalHeader +
+                                 pe.FILE_HEADER.NumberOfSections * _SECTION_HEADER_SIZE)
+            if 'header_space' not in test and pe.OPTIONAL_HEADER.SizeOfHeaders - section_table_end >= _SECTION_HEADER_SIZE:
+                # there's enough unused space to add new section header
+                new_section_header_space_needed = 0
+            else:
+                # There's not enough unused space to add new section header;
+                # calculate how much extra space to add. In nearly all cases,
+                # FileAlignment extra bytes will be added.
+                new_section_header_space_needed = _round_to_next(_SECTION_HEADER_SIZE, pe.OPTIONAL_HEADER.FileAlignment)
+            new_section_data_size = len(new_section_data)
+            new_section_data_padded_size = _round_to_next(new_section_data_size, pe.OPTIONAL_HEADER.FileAlignment)
+            new_section_rva = _round_to_next(max(section.VirtualAddress + section.Misc_VirtualSize for section in pe.sections), pe.OPTIONAL_HEADER.SectionAlignment)
+
+            pe.FILE_HEADER.NumberOfSections += 1
+            pe.OPTIONAL_HEADER.SizeOfInitializedData += new_section_data_padded_size
+            pe.OPTIONAL_HEADER.SizeOfImage = _round_to_next(new_section_rva + new_section_data_size, pe.OPTIONAL_HEADER.SectionAlignment)
+            pe.OPTIONAL_HEADER.SizeOfHeaders += new_section_header_space_needed
+            for section in pe.sections:
+                section.PointerToRawData += new_section_header_space_needed
+
+            # update import tables
+            if hasattr(pe, 'DIRECTORY_ENTRY_IMPORT'):
+                for entry in pe.DIRECTORY_ENTRY_IMPORT:
+                    old_dll = entry.dll.lower()
+                    if old_dll in new_section_offset_mapping:
+                        entry.struct.Name = new_section_rva + new_section_offset_mapping[old_dll]
+            if hasattr(pe, 'DIRECTORY_ENTRY_DELAY_IMPORT'):
+                for entry in pe.DIRECTORY_ENTRY_DELAY_IMPORT:
+                    old_dll = entry.dll.lower()
+                    if old_dll in new_section_offset_mapping:
+                        entry.struct.szName = new_section_rva + new_section_offset_mapping[old_dll]
+
+        # clear any signatures
         cert_table = pe.OPTIONAL_HEADER.DATA_DIRECTORY[pefile.DIRECTORY_ENTRY['IMAGE_DIRECTORY_ENTRY_SECURITY']]
-        if cert_table.Size > 0:
+        if cert_table.Size:
             warnings.warn(f'Authenticode signature has been removed from {lib_name}', UserWarning)
             cert_table.VirtualAddress = 0
             cert_table.Size = 0
 
-        if hasattr(pe, 'DIRECTORY_ENTRY_IMPORT'):
-            for entry in pe.DIRECTORY_ENTRY_IMPORT:
-                old_dll = entry.dll.lower()
-                if old_dll in section_offset_mapping:
-                    entry.struct.Name = new_section_rva + section_offset_mapping[old_dll]
-        if hasattr(pe, 'DIRECTORY_ENTRY_DELAY_IMPORT'):
-            for entry in pe.DIRECTORY_ENTRY_DELAY_IMPORT:
-                old_dll = entry.dll.lower()
-                if old_dll in section_offset_mapping:
-                    entry.struct.szName = new_section_rva + section_offset_mapping[old_dll]
+        # all changes to headers are done; serialize the PE file
         lib_data = pe.write()
 
-    # add new section header and new section
-    with io.BytesIO() as new_lib_data:
-        new_lib_data.write(lib_data[:section_table_end])
-        new_section_header = struct.pack(
-            _SECTION_HEADER_FORMAT,
-            b'redll',  # Name
-            len(new_section_data),  # VirtualSize
-            new_section_rva,  # VirtualAddress
-            new_section_data_size,  # SizeOfRawData
-            len(lib_data) + new_section_header_space_needed,  # PointerToRawData
-            0,  # PointerToRelocations
-            0,  # PointerToLinenumbers
-            0,  # NumberOfRelocations
-            0,  # NumberOfLinenumbers
-            pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_MEM_READ'] | pefile.SECTION_CHARACTERISTICS['IMAGE_SCN_CNT_INITIALIZED_DATA']
-                # Characteristics
-        )
-        new_lib_data.write(new_section_header)
-        if new_section_header_space_needed:
-            new_lib_data.write(b'\x00' * (new_section_header_space_needed - _SECTION_HEADER_SIZE))
-            new_lib_data.write(lib_data[section_table_end:])
-        else:
-            new_lib_data.write(lib_data[section_table_end + _SECTION_HEADER_SIZE:])
-        new_lib_data.write(new_section_data)
-        new_lib_data.write(b'\x00' * (new_section_data_size - len(new_section_data)))
-        lib_data = new_lib_data.getvalue()
+        if not enough_padding:
+            # add new section header and new section
+            with io.BytesIO() as new_lib_data:
+                new_lib_data.write(lib_data[:section_table_end])
+                new_section_header = struct.pack(
+                    _SECTION_HEADER_FORMAT,
+                    b'dlvwhl',  # Name
+                    new_section_data_size,  # VirtualSize
+                    new_section_rva,  # VirtualAddress
+                    new_section_data_padded_size,  # SizeOfRawData
+                    len(lib_data) + new_section_header_space_needed,  # PointerToRawData
+                    0,  # PointerToRelocations
+                    0,  # PointerToLinenumbers
+                    0,  # NumberOfRelocations
+                    0,  # NumberOfLinenumbers
+                    _NEW_SECTION_CHARACTERISTICS  # Characteristics
+                )
+                new_lib_data.write(new_section_header)
+                if new_section_header_space_needed:
+                    new_lib_data.write(b'\x00' * (new_section_header_space_needed - _SECTION_HEADER_SIZE))
+                    new_lib_data.write(lib_data[section_table_end:])
+                else:
+                    new_lib_data.write(lib_data[section_table_end + _SECTION_HEADER_SIZE:])
+                new_lib_data.write(new_section_data)
+                new_lib_data.write(b'\x00' * (new_section_data_padded_size - new_section_data_size))
+                lib_data = new_lib_data.getvalue()
 
     # fix the checksum
     with PEContext(None, lib_data, False, verbose) as pe:
